/* tga.c
 * Fonctions permettant la lecture et l'criture d'images Targa TGA
 *
 * Auteur: VOURIOT Thierry
 * Web: http://thierry.vouriot.free.fr
 * Email: thierry.vouriot@free.fr
 * Date: 09 septembre 2002
 */
 
/* log
 * 30/08/2002 prise en charge des images 8 bits
 * 09/09/2002 prise en charge des images 16 bits
 */
 
/*
 *
 * Le format Targa TGA est divis en 2 partie l'en-tte et les donnes.
 * L'en-tte contient 12 champs et est dfinie par la structure suivante :
 *
 * typedef struct tagTARGAFILEHEADER
 * {
 *	unsigned char imageIDLength;	// number of characters in identification field;
 *					// 0 denotes no identification field is included
 *	unsigned char colorMapType;	// type of color map; always 0
 *	unsigned char imageTypeCode;	// uncompressed RGB is 2;
 *					// uncompressed grayscale is 3
 *	short int colorMapOrigin;	// origin of color map (lo-hi); always 0
 *	short int colorMapLength;	// length of color map (lo-hi); always 0
 *	short int colorMapEntrySize;	// color map entry size (lo-hi); always 0
 *	short int imageXOrigin;		// x coordinate of lower-left corner of image
 *					// (lo-hi); always 0
 *	short int imageYOrigin;		// y coordinate of lower-left corner of image
 *					// (lo-hi); always 0
 *	short int imageWidth;		// width of image in pixels (lo-hi)
 *	short int imageHeight;		// height of image in pixels (lo-hi)
 *	unsigned char bitCount;		// number of bits; 16, 24, 32
 *	unsigned char imageDescriptor;	// 24 bit = 0x00; 32 bit = 0x08
 * } TARGAFILEHEADER
 *
 */

#include "tga.h"

/* Fonction permettant de tester si un entier est une puisance de 2 */
int checkSize(int x)
{
    if (x == 2	 || x == 4 || 
        x == 8	 || x == 16 || 
        x == 32  || x == 64 ||
        x == 128 || x == 256 || 
	x == 512 || x == 1024)
        return 1;
    else return 0;
}

/* Fonction lisant un fichier TGA 8, 16, 24 ou 32 bits et remplissant la structure
 * TGAFILE passe en paramtre
 */
int LoadTGAFile(char *filename, TGAFILE *tgaFile) 
{
	FILE *filePtr;			// pointeur sur le fichier
	unsigned char ucharBad;		// poubelle pour les unsigned char
	short int sintBad;		// poubelle pour les short int
	int colorMode;			// 4 pour RGBA ou 3 pour RGB
	long imageSize;			// taille de l'image TGA
	long imageIdx;			// compteur
	unsigned char colorSwap;	// variable d'change
	unsigned short pixels = 0;	// un pixel d'une image 16 bits
	int r, g, b;			// composantes RGB
	
	// ouverture du fichier TGA
	if (!(filePtr = fopen (filename, "rb"))) {
		fclose(filePtr);
        	return TGA_FILE_NOT_FOUND;
	}
	
	// lecture des 2 premiers champs dont on n'a pas besoin
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	
	// lecture du type de l'image
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	if ((ucharBad != 2) && (ucharBad != 3)) {
		fclose(filePtr);
		return TGA_BAD_IMAGE_TYPE;
	}
	// lecture des 5 champs suivants dont on n'a pas besoin
	fread(&sintBad, sizeof(short int), 1, filePtr);
	fread(&sintBad, sizeof(short int), 1, filePtr);
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	fread(&sintBad, sizeof(short int), 1, filePtr);
	fread(&sintBad, sizeof(short int), 1, filePtr);
	
	// lecture des dimensions de l'image
	fread(&tgaFile->imageWidth, sizeof(short int), 1, filePtr);
	fread(&tgaFile->imageHeight, sizeof(short int), 1, filePtr);
	fread(&tgaFile->bitCount, sizeof(unsigned char), 1, filePtr);
	
	// lecture du dernier champ dont on n'a pas besoin
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	
	// on ne lit que les images 24 ou 32 bits
	if (tgaFile->bitCount != 32 && tgaFile->bitCount != 24 && tgaFile->bitCount != 8 && tgaFile->bitCount != 16) {
		fclose(filePtr);
		return TGA_BAD_BITS;
	}
	
	// image 8, 24 ou 32 bits
	if(tgaFile->bitCount == 24 || tgaFile->bitCount == 32 || tgaFile->bitCount == 8)
	{
		// colorMode -> 3 = BGR, 4 = BGRA
		colorMode = tgaFile->bitCount / 8;
		imageSize = tgaFile->imageWidth * tgaFile->imageHeight * colorMode;
	
		// allocation de la mmoire pour les donnes de l'image
		tgaFile->imageData = (unsigned char*) malloc(sizeof(unsigned char) * imageSize);
	
		// lecture des donnes de l'image
		fread(tgaFile->imageData, sizeof(unsigned char), imageSize, filePtr);
		if (tgaFile->imageData == NULL) {
        		fclose(filePtr);
			return TGA_BAD_DATA;
		}
	
		// dans les images TGA les couleurs des pixels sont en mode BGR
		// transformation en mode RGB
		if (tgaFile->bitCount != 8)
		for (imageIdx=0; imageIdx < imageSize; imageIdx += colorMode)
		{
			colorSwap = tgaFile->imageData[imageIdx];
			tgaFile->imageData[imageIdx] = tgaFile->imageData[imageIdx + 2];
			tgaFile->imageData[imageIdx + 2] = colorSwap;
		}
	}
	else { // image 16 bits

		colorMode = 3;	// mode RGB
		imageSize = tgaFile->imageWidth * tgaFile->imageHeight * colorMode;
		
		// allocation de la mmoire pour les donnes de l'image
		tgaFile->imageData = (unsigned char*) malloc(sizeof(unsigned char) * imageSize);
		
		// en 16 bits chaque pixel est contenu dans un "unsigned short"
		// boucle sur tous les pixels de l'image
		for (imageIdx = 0; imageIdx < tgaFile->imageWidth*tgaFile->imageHeight; imageIdx++)
		{
			// lecture du pixel courant
			fread(&pixels, sizeof(unsigned short), 1, filePtr);
				
			// pour convertir 16 bits en composantes RGB
			// chaque composante occupe 5 bits
			// ex: pixels = 0x61E8 = 0110000111101000
			// on utilise un masque 0x1F = 11111
			// 0110000111101000 & 11111 = 01000
			// on effectue un dcallage de 3 bits pour avoir un nombre sur 8 bits
			// 01000 << 3 = 01000000 = 0x40 = 64 --> composante bleu
			// on dcale ensuite de 5 bits pour arriver  la prochaine
			// composante et on ritre la mme opration
			// 0110000111101000 >> 5 = 0000001100001111
			// 0000001100001111 & 11111 = 01111
			// 01111 << 3 = 01111000 = 0x78 = 120 --> composante vert
			// ...etc...
			b = (pixels & 0x1f) << 3;
			g = ((pixels >> 5) & 0x1f) << 3;
			r = ((pixels >> 10) & 0x1f) << 3;
				
			// on assigne ensuite les composantes au tableau de donnes
			tgaFile->imageData[imageIdx * 3 + 0] = r;
			tgaFile->imageData[imageIdx * 3 + 1] = g;
			tgaFile->imageData[imageIdx * 3 + 2] = b;
		}
	}
	
	// fermeture du fichier
	fclose(filePtr);
	
	// Format de l'image RGB, RGBA ou niveau de gris
	switch (tgaFile->bitCount) {
		case 8: tgaFile->texFormat = GL_LUMINANCE; break;
		case 32: tgaFile->texFormat = GL_RGBA; break;
		default: tgaFile->texFormat = GL_RGB;
	}
	
	return 1;
}

/* Fonction crant une image TGA 24 bits dans le fichier filename */
int Write24bitsTGAFile(char *filename, short int width, short int height, unsigned char* imageData)
{
	unsigned char byteSkip;		// utilis pour mettre  0 certains champs
	short int shortSkip;		// utilis pour mettre  0 certains champs
	unsigned char imageType;	// type de l'image
	int colorMode;			// color mode
	unsigned char colorSwap;	// utilis pour la conversion RGB - BGR
	int imageIdx;			// compteur
	unsigned char bitDepth;		// image 24 bits
	long imageSize;			// taille de l'image
	FILE *filePtr;			// pointeur sur le fichier
	
	// cration du fichier en mode criture binaire
	if (!(filePtr = fopen (filename, "wb"))) {
		fclose(filePtr);
        	return TGA_FILE_NOT_CREATE;
	}
	
	imageType = 2;	// RGB uncompressed
	bitDepth = 24;	// 24 bitdepth
	colorMode = 3;	// RGB color mode
	
	byteSkip = 0;
	shortSkip = 0;
	
	// criture des 2 premiers champs  0
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	
	// criture du type de l'image
	fwrite(&imageType, sizeof(unsigned char), 1, filePtr);
	
	// criture des 5 champs suivants  0
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	
	// criture des dimensions de l'image
	fwrite(&width, sizeof(short int), 1, filePtr);
	fwrite(&height, sizeof(short int), 1, filePtr);
	fwrite(&bitDepth, sizeof(unsigned char), 1, filePtr);
	
	// criture du dernier champ  0
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	
	// calcul de la taille de l'image
	imageSize = width * height * colorMode;
	
	// conversion des donnes du mode RGB vers le mode BGR
	for (imageIdx = 0; imageIdx < imageSize ; imageIdx += colorMode)
	{
		colorSwap = imageData[imageIdx];
		imageData[imageIdx] = imageData[imageIdx + 2];
		imageData[imageIdx + 2] = colorSwap;
	}
	
	// criture des donnes de l'image dans le fichier
	fwrite(imageData, sizeof(unsigned char), imageSize, filePtr);
	
	// fermeture du fichier
	fclose(filePtr);
	
	return 1;	
}

